from .list import List_Models
from .clone import Clone_Model
from .config import Config
from .remove import Remove_Models
from .run import Run_Model
from .upload import Upload_Model
from .install import Install_Backend

__all__ = [
    "List_Models",
    "Clone_Model",
    "Config",
    "Remove_Models",
    "Run_Model",
    "Upload_Model",
    "Install_Backend",
]


from PIserver.clients.FileDownloadClient import FileDownloadClient
from PIserver.clients.LLMClient import *
from PIserver.commands.command import Command
from PIserver.constants import *
from pathlib import Path
from PIserver.install import local_compile
from PIserver.setup import set_up
from PIserver.utils.files import *
import subprocess
from tqdm import tqdm
import time
import keyboard

class Run_Model(Command):
    def __init__(self):
        super().__init__()
        self.state = LLMState.STOPPED
        
    def register_subcommand(self, subparser):
        run_parser = subparser.add_parser("run", help="Run a large language model.")
        run_parser.add_argument("model", help="The model to run. Can be a local model path or a remote model name.")
        run_parser.add_argument("-cfg","--config", default=None, help="The configuration file to use.")
        run_parser.add_argument("--no-update", action="store_true", default=None, help="Do not update the model if it is outdated.")
        
    def execute(self, args):
        if not Path(DEFAULT_STORAGE_PATH).exists():
            set_up()
        # check config file
        cfg = self.check_config(args.config)

        # check backend engine
        if 'engine' not in cfg:
            engine = local_compile()
            
            if engine is not None:
                add_engine(DEFAULT_ENGINE_NAME, engine)
                cfg['engine'] = DEFAULT_ENGINE_NAME
                write_file(DEFAULT_CONFIG_FILE, cfg)
                print("Engine successfully installed.")
            else:
                log_error("Unable to compile engine. \
                    Please compile the code from PowerInfer repo https://github.com/SJTU-IPADS/PowerInfer manually and install it in order to use.")
                return
        else:    
            engine = self.check_engine(cfg['engine'])
            
        if engine is None:
            log_error(f"Unable to find {engine} in the installed list. Trying to install remote engine and find local files...")
            # check local path
            engine = cfg['engine']
            if not Path(engine).exists():
                log_error(f"Unable to find {engine} locally. Please install the engine first.")
                return
                
        
        # check model existence
        mname = str(args.model)
        if ':' not in mname or '/' not in mname: 
            log_error("Invalid model name. Format the model name like 'USR/NAME:SIZE'.")
            return
        row, rest = filter_rows(parse_condition(mname))
        mpath = None
        if len(row) == 0:
            # check if it is local model dir
            mpath = mname if check_existence(mname) else None
        else:
            print("Found model locally.")
            model_info = row[0]
            mpath = model_info[4] if check_existence(model_info[4]) else None
        if mpath is None or args.no_update is None:
            # download model from remote
            client = FileDownloadClient()
            local_dir_name = str(args.model).replace(":", "-").replace("/", "-")
            local_path = Path(read_file(DEFAULT_CONFIG_FILE)["model_path"]) / Path(local_dir_name)
            local_path.mkdir(parents=True, exist_ok=True)
            mname, tname = check_model_name_with_size(args.model)
            mname, uname = get_uname_from_model(mname)
            if tname == "":
                log_error("Please specify the model size. Format the model name like 'USR/NAME:SIZE'.")
                return
            getModelRequest = {
                "mname": mname,
                "tname": tname
            }
            if uname!= "":
                getModelRequest["uname"] = uname
            
            try:
                info = client.getModelInfo(getModelRequest)
                if info is None:
                    return 
                if mpath is not None and row[0][3] == info["version"]:
                    print(f"Model {mname} is up to date.")
                else:
                    print(f"Model {mname} is outdated. Downloading new version...")
                    print(f"Downloading model {mname} from remote...")
                    if client.download(local_path, info["dir_info"]):
                        name = uname + "/" + mname if uname != "" else mname
                        add_row([name, tname, info["size"], info["version"], str(local_path)])
                        mpath = str(local_path)
            except KeyboardInterrupt:
                print("Download stopped.")
                return
            
        # find gguf model weight, if multiple, use first one found
        mpath = Path(mpath)
        if mpath.is_dir():
            model_files = list(mpath.glob("*.gguf"))
            if len(model_files) == 0:
                log_error(f"Unable to find model weight file in {mpath}. Please check the model directory.")
                return
            elif len(model_files) > 1:
                print(f"Multiple model weight files detected. Using `powerinfer run local_model_path` to specify what you need. Currently using {str(model_files[0])}.")
            mpath = model_files[0]
        
        # format command
        cmd = f"{engine} -m {str(mpath)} -np 4"
        for option in cfg:
            if self.filter_options(option):
                cmd += f" --{option} {cfg[option] if type(cfg[option]) is not bool else ''}"
                
        print(f"Running model with command: {cmd}")
        
        process = subprocess.Popen(
            cmd.split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        time.sleep(1)
        if process.poll() is not None:
            out, err = process.communicate()
            log_error(f"Unable to start the model service. {out} {err}")
            log_error("This could possibly be caused by unsupported model architecture or invalid engine. \
                Please check your engine or leave an issure in the github respository.")
            return
        print("Start to load model...")
        pbar = tqdm(total=100, desc="Loading model", unit="%")
        
        while True:
            output = str(process.stdout.readline())
            err = process.stderr.readline() if process.stderr is not None else ''
            if err != '':
                log_error(err)
                return
            
            if pbar.n < 13 and "llama_model_loader" in output:
                pbar.update(13-pbar.n)
            if pbar.n < 33 and "kv" in output:
                pbar.update(33-pbar.n)
            if pbar.n < 40 and "llama_model_load" in output:
                pbar.update(40-pbar.n)
            if pbar.n < 52 and "llama_new_context" in output:
                pbar.update(52-pbar.n)
            if pbar.n < 71 and "llama_build_graph" in output:
                pbar.update(71-pbar.n)
            if "llama server listening at" in output:
                pbar.update(100-pbar.n)
                pbar.close()
                print("Model successfully loaded.")
                break
            # print(output)
        
        
        print()
        print("Pressing 'CTRL+D' to stop inferencing.")
        print()
        prompt_manager = PromptManager(cfg['system-prompt'] if 'system-prompt' in cfg else '')
        keyboard.add_hotkey("ctrl+d", self.set_stop, suppress=True)
            
        while True:
            try:
                prompt = input(">>> ")
                self.state = LLMState.RUNNING
                params = {
                    "prompt": prompt_manager.format_prompt(prompt),
                    "stream": True,
                }
                if 'options' in cfg:
                    params.update(cfg["options"])
                
                client = LLMClient(POWERINFER_LOCAL_MODEL_HOST)
                answer = ""
                for chunk in client.generate(params):
                    if self.state == LLMState.STOPPED:
                        break
                    data = json.loads(chunk)
                    if 'choices' in data and len(data['choices']) > 0:
                        if 'delta' in data['choices'][0] and 'content' in data['choices'][0]['delta']:
                            content = data['choices'][0]['delta']['content']
                            print(content, end="", flush=True)
                            answer += content
                
                print()            
                prompt_manager.save_dialog(prompt, answer)

            except json.JSONDecodeError:
                log_error("Unable to decode json from server.")
                break
            except KeyboardInterrupt:
                print()
                print("Trying to stop model service...")
                process.terminate()
                process.wait()
                print("Model service successfully stopped.")
                break
            except Exception as e:
                log_error(e)
                process.terminate()
                process.wait()
                break
            
        return
        
        
    def check_engine(self, name):
        engines = read_file(DEFAULT_ENGINE_LIST_FILE)
        if name not in engines:
            log_error(f"Engine {name} not found. Please install it using `powerinfer install` first.")
            return None
        return engines[name]
    
    def check_config(self, file):
        cfg_file = DEFAULT_CONFIG_FILE
        if file is None:
            print("Running model using default configuration...")
        else:
            cfg_file = Path(file)
            print(f"Running model using configuration file {file}...")
        if not cfg_file.exists():
            print(f"Configuration file {file} not found. Creating configuration file with default config options...")
        cfg = read_file(cfg_file)
        if len(cfg) == 0:
            cfg = DEFAULT_CONFIG
            write_file(cfg_file, cfg)
        return cfg
    
    def filter_options(self, option: str):
        anti = ["model_path", "engine", "options", "system-prompt"]
        for a in anti:
            if a == option:
                return False
        return True
    
    def set_stop(self):
        if self.state == LLMState.RUNNING:
            print()
            print("Stopping signal sent. Inference stopped.")
            
            self.state = LLMState.STOPPED


import json
from tqdm import tqdm
import os
from pathlib import Path
from PIserver.clients.net import send_post_request
from PIserver.utils.files import log_error
from PIserver.constants import CHUNK_SIZE

class FileDownloadClient:
    def __init__(self):
        pass
    
    def getModelInfo(self, body):
        res = send_post_request("/type/client/get", body)
        
        if res is None:
            log_error("Cannot get response from server. Please check your internet connection.")
            return None
        if res.status_code != 200:
            log_error(f"Server returned {res.status_code} - {res.text}")
            return None
        
        res = dict(res.json())
        if res["state"] == "SUCCESS":
            return res["model"]
        else:
            log_error(f"{res['state']} - {res['message']}")
            
    def download_file(self, local_path, remote_path):
        headers = {}
        local_file = Path(local_path)
        if local_file.exists():
            downloaded_size = local_file.stat().st_size
            headers["Range"] = f"bytes={downloaded_size}-"
        else:
            downloaded_size = 0

        try:
            response = send_post_request("/type/download",params={"path": remote_path}, header=headers, stream=True)
            if response is None:
                log_error("Cannot get response from server. Please check your internet connection.")
                return False
            if response.status_code == 404:
                log_error(f"File not found: {remote_path}")
                return False
            response.raise_for_status()

            content_range = response.headers.get("Content-Range", "")
            if content_range:
                total_size = int(content_range.split("/")[-1])
            else:
                total_size = int(response.headers.get("Content-Length", 0)) + downloaded_size

            mode = "ab" if downloaded_size > 0 else "wb"
            with open(local_path, mode) as file, tqdm(
                total=total_size, 
                unit='B', 
                unit_scale=True, 
                desc=os.path.basename(local_path),
                initial=downloaded_size
            ) as pbar:
                for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                    file.write(chunk)
                    pbar.update(len(chunk))
            return True
        except Exception as e:
            log_error(f"Failed to download file {remote_path}: {e}")
            return False

    def iter_folder(self, local_path: Path, files: dict):
        for name, fpath in files.items():
            if isinstance(fpath, str):
                file_path = local_path / Path(name)
                if not self.download_file(file_path, fpath):
                    continue  # Skip to next file if download fails
            else:
                subfolder_path = local_path / Path(name)
                subfolder_path.mkdir(parents=True, exist_ok=True)
                self.iter_folder(subfolder_path, fpath)
            
    def download(self, local_path: Path, structure: str):
        try:
            files = dict(json.loads(structure))
            # print("parsing files=======>",files) 
            local_path.mkdir(parents=True, exist_ok=True)

            self.iter_folder(local_path, files)

            print(f"Model successfully cloned into {str(local_path.absolute())}")
            return True
        except Exception as e:
            log_error(f"Failed to clone file: {e}")
            return False


from PIserver.commands.command import Command
from PIserver.install import local_compile
from PIserver.setup import set_up
from PIserver.utils.install import *
from PIserver.constants import DEFAULT_ENGINE_NAME
from PIserver.utils.files import *

class Install_Backend(Command):
    def register_subcommand(self, subparser):
        install_parser = subparser.add_parser("install", help="Install the backend engine.")
        install_parser.add_argument("engine", nargs='?', default=None, help="The backend engine name to install.")
        install_parser.add_argument("-f","--file", help="The local runnable file compiled by yourself from powerinfer.")
        install_parser.add_argument("-c","--compile", help="Download the source code and compile it automatically. Require c compilers and cmake installed.")
        
    def execute(self, args):
        if not Path(DEFAULT_STORAGE_PATH).exists():
            set_up()
        engine_name = None
        if args.file is not None and args.engine is None:
            log_error("Please specify the engine name in order to use.")
            return
        if args.compile is not None:
            engine_path = local_compile()
            if engine_path is None:
                log_error("Unable to compile the engine. Please check the source code and cmake and try again.")
                log_error("If you have already manually compiled the engine, please use -f to install it.")
                return
            res = add_engine(
                args.engine if args.engine is not None else DEFAULT_ENGINE_NAME,
                engine_path
            )
            if not res:
                return
            
            engine_name = DEFAULT_ENGINE_NAME
            
        if args.engine is not None:
            engine_name = args.engine
            if args.file is None:
                if args.engine not in list(get_engine_choices().keys()):
                    log_error(f"Engine {args.engine} not found. Please check the name or use `powerinfer list -i` to see all the available engines or install engine compiled by yourself using -f.")
                else:
                    remote_path = get_engine_path(args.engine)
                    res = add_engine(args.engine, single_install(remote_path))
                    if not res:
                        return         
            else:
                # install from file
                res = add_engine(args.engine, args.file)
                if not res:
                    return
        else:
            interactive_install()
            
        if engine_name is not None:
            config = read_file(DEFAULT_CONFIG_FILE)
            config["engine"] = engine_name
            write_file(DEFAULT_CONFIG_FILE, config)
            
        print("Engine successfully installed.")
        
    


import argparse
from pkg_resources import get_distribution
from PIserver.constants import *
from PIserver.commands import *

        
def get_version():
    return get_distribution("powerinfer-server").version 

command_map = {
    "list": List_Models(),
    "remove": Remove_Models(),
    "clone": Clone_Model(),
    "config": Config(),
    "upload": Upload_Model(),
    "run": Run_Model(),
    "install": Install_Backend(),
}
    
def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("-v","--version",help="show version", action="store_true")
    
    subparsers = parser.add_subparsers(dest="command")
    
    for cmd in command_map:
        command_map[cmd].register_subcommand(subparsers) 
    
    return parser
        
def main():
    parser = create_parser()
    args = parser.parse_args()
    
    if args.version:
        print(f"powerinfer-server version {get_version()}")
        return

    if args.command is None or args.command not in command_map:
        parser.print_help()
        return
    
    command_map[args.command].execute(args)
    

if __name__ == "__main__":
    main()    


from datetime import datetime
from pathlib import Path
from PIserver.clients.FileUploadClient import FileUploadClient
from PIserver.clients.net import send_post_request
from PIserver.commands.command import Command
from PIserver.constants import DEFAULT_STORAGE_PATH
from PIserver.setup import set_up
from PIserver.utils.files import log_error


class Upload_Model(Command):
    def register_subcommand(self, subparser):
        upload_parser = subparser.add_parser("upload", help="Upload a local model to https://powerinfer.com. and get predictors.")
        upload_parser.add_argument("model", help="The remote model repository to upload.")
        upload_parser.add_argument("-d","--local-dir", default=None, help="The local model directory to upload.")
        upload_parser.add_argument("-hf", "--huggingface", default=None, help="Upload the model from huggingface hub.")
        upload_parser.add_argument("-s", "--status", default=None, help="Query current model training process status.", action="store_true")
        upload_parser.add_argument("-c", "--cancel", default=None, help="Cancel existing training process.", action="store_true")
        upload_parser.add_argument("--no-train", default=False, help="Upload a model without training.", action="store_true")

    def execute(self, args):
        if not Path(DEFAULT_STORAGE_PATH).exists():
            set_up()
        if ':' not in str(args.model):
            log_error("Invalid model name. Please specify the size in the format NAME:SIZE.")
            return
        if args.status is not None:
            response = send_post_request("/task/client/detail", params={
                "mname": str(args.model).split(":")[0],
                "tname": str(args.model).split(":")[1]
            })
            if response is None:
                log_error("Cannot successfully get response. Please check your internet connection.")
                return
            if response.status_code == 404:
                log_error(f"Model {args.model} not found.")
                return
            task = dict(response.json())
            # {'tid': 'e8ec8383da200308f2f83fbfdf57589a', 'dir': 'D://train/73027646d141476489fd6dae05bdb217/testLlama-7b', 'version': 'd83d625b0d7683711c00bbd035eed394ea141609', 'id': 'c0ecd6558ad2e0838663e8b7b2c9bf9e', 'state': 'UPLOADING', 'created': '2025-03-27T20:56:30.123858', 'started': None, 'finished': None, 'queue': 0, 'progress': 0, 'waitingTime': 0, 'runningTime': 0}
            print("STATE: ", task["state"])
            print("version: ", task["version"])
            print("Task created at: ", datetime.fromisoformat(task["created"]).astimezone().strftime("%Y-%m-%d %H:%M:%S %Z%z"))
            print("Task started at: ", datetime.fromisoformat(task["started"]).astimezone().strftime("%Y-%m-%d %H:%M:%S %Z%z") if task["started"] is not None else "--")
            print("Task finished at: ", datetime.fromisoformat(task["finished"]).astimezone().strftime("%Y-%m-%d %H:%M:%S %Z%z") if task["finished"] is not None else "--")
            
            return
        
        client = FileUploadClient(str(args.model).split(":"))
        if args.cancel is not None:
            client.cancel()
            return 
        
        try:
            print("Uploading model... Press Ctrl+C to cancel task.")
            if args.huggingface is not None:
                client.upload(args.local_dir ,no_train=args.no_train, hf=args.huggingface)
                return
            client.upload(args.local_dir, no_train=args.no_train)
             
        except KeyboardInterrupt:
            client.cancel()
            return
        


import hashlib
import json
from PIserver.clients.net import getHeader, send_post_request
from PIserver.constants import *
from pathlib import Path
import requests
from tqdm import tqdm

from PIserver.utils.files import log_error
import mmap

class FileUploadClient():
    def __init__(self, model_name: list):
        self.host = POWERINFER_MODEL_HOST
        self.port = POWERINFER_SERVER_PORT
        self.mname = model_name[0]
        self.tname = model_name[1]
        
    def generate_md5(self, file_path: Path):
        print(f"Validating file changes {str(file_path)}...")
        md5_hash = hashlib.md5()
        buffer_size = 1024 * 1024  # 1MB
        
        with open(file_path, "rb") as f:
            with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mmapped_file:
                while chunk := mmapped_file.read(buffer_size):
                    md5_hash.update(chunk)
        return md5_hash.hexdigest()


    def upload_file(self, path: Path, need_train: bool) -> bool:
        file_size = path.stat().st_size
        uploaded_size = 0
        auth_header = getHeader()
        response = requests.head(
            backend_host+"/task/client/upload",  
            headers=auth_header,
            data=json.dumps({
                "mname": self.mname, 
                "tname": self.tname, 
                "fname": path.name, 
                "md5": self.generate_md5(path),
                "need_train": need_train
            })
        )
        if response.status_code == 403:
            print(response.text)
            return False
        if 'Content-Range' in response.headers:
            uploaded_size = int(response.headers['Content-Range'].split('/')[1])
            
        if uploaded_size == file_size:
            print(f"File {path.name} is already uploaded.")
            return True
        
        with tqdm(total=file_size, unit='B', unit_scale=True, desc=path.name, initial=uploaded_size) as pbar:
            with open(path, 'rb') as f:
                f.seek(uploaded_size)
                while uploaded_size < file_size:
                    chunk = f.read(CHUNK_SIZE)
                    if not chunk:
                        break
                    
                    headers = {
                        'Content-Range': f'bytes {uploaded_size}-{uploaded_size + len(chunk) - 1}/{file_size}'
                    }
                    headers.update(auth_header)
                    response = requests.patch(
                        backend_host+"/task/client/upload", 
                        headers=headers, 
                        params={
                            "mname": self.mname, 
                            "tname": self.tname, 
                            "fname": path.name,
                            "need_train": need_train
                        }, 
                        data=chunk
                    )
                    
                    if response.status_code == 200 or response.status_code == 206:
                        uploaded_size += len(chunk)
                        pbar.update(len(chunk))
                    else:
                        log_error(f"Failed to Upload. {response.text}")
                        return False
        return True
    
    def iter_folder(self, path: Path, need_train: bool):
        print(f"Iterating folder {str(path)}...")
        success = True 
        if path.is_file() and not path.name.startswith("."):
            success = self.upload_file(path, need_train) and success
        elif path.is_dir() and not path.name.startswith("."):
            for file in path.iterdir():
                if file.is_file() and not file.name.startswith("."):
                    success = self.upload_file(file, need_train) and success
                elif file.is_dir() and not file.name.startswith("."):
                    success = self.iter_folder(file, need_train) and success
        return success
                    
    def upload(self, file_path, no_train: bool = False, hf: str = ""):
        if hf == "" and not Path(file_path).exists():
            log_error(f"Folder {file_path} does not exist.")
            return
        
        params = {
            "mname": self.mname,
            "tname": self.tname,
            "need_train": not no_train
        }
        if hf is not None and hf != "":
            params["hf_path"] = hf
        response = send_post_request("/task/client/add", params=params)
        if response is None:
            log_error("Cannot successfully get response. Please check your internet connection.")
            return
        
        if response.status_code == 403:
            log_error(response.text)
            return
        if response.status_code != 200:
            log_error(f"Failed to upload. Please try it later.")
            return
        
        if hf is not None and hf != "":
            print(f"Uploading huggingface model {hf} to {self.mname+':'+self.tname} task submitted successfully.")
            print(f"Use `powerinfer upload {self.mname+':'+self.tname} -s` to check the process.")
            return
        
        success = self.iter_folder(Path(file_path), need_train=not no_train)
        
        send_post_request("/task/client/done", params={"mname": self.mname, "tname": self.tname, "success": success})
        if success:      
            print(f"Upload successfully. Visit xxx.com or use `powerinfer upload {self.mname+':'+self.tname} -s` to check the process.")
            print(f"Use `powerinfer upload {self.mname+':'+self.tname} -c` to cancel the training process.")
        else:
            log_error("Upload failed. Please rerun the command to resume uploading model.")
            return
    
    # def fetch_history(self):
    #     response = send_post_request("/task/client/query", params={"mname": self.mname, "tname": self.tname}, stream=True)
    #     if response.status_code == 200:
    #         try:
    #             for line in response.iter_lines():
    #                 if line:
    #                     task = json.loads(line)
    #                     yield task
    #         except Exception as e:
    #             log_error(f"Failed to parse task stream: {e}")
    #             yield None
        
    def cancel(self):
        print("Starting to cancel training task for model", self.mname + ':' + self.tname)
        response = send_post_request("/task/client/cancel", params={"mname": self.mname, "tname": self.tname})
        if response is None:
            log_error("Cannot successfully get response. Please check your internet connection.")
            return
        if response.text == "true":
            print("Cancelled successfully.")
        else:
            print("No running task found.")
        


import os
import subprocess
import zipfile
from PIserver.constants import *
import requests

def download_source_code():
    source_url = "https://github.com/SJTU-IPADS/PowerInfer/archive/refs/heads/main.zip"
    tmp_path = DEFAULT_STORAGE_PATH / "main.zip"
    
    try:
        print(f"Downloading source code from {source_url} to {tmp_path}...")
        response = requests.get(source_url, stream=True)
        response.raise_for_status()
        
        with open(tmp_path, 'wb') as file:
            for chunk in response.iter_content(chunk_size=8192):
                file.write(chunk)

        print("Download completed.")
        with zipfile.ZipFile(tmp_path, 'r') as zip_ref:
            zip_ref.extractall(DEFAULT_STORAGE_PATH)
    
    except requests.exceptions.RequestException as e:
        print(f"Error downloading source code: {e}")
        return None
    except zipfile.BadZipFile as e:
        print(f"Error extracting ZIP file: {e}")
        return None
    finally:
        if tmp_path.exists():
            os.remove(tmp_path)
        return DEFAULT_STORAGE_PATH / "PowerInfer-main"


def local_compile() -> str:
    '''
    Download the source code and compile it.
    Return the path of the compiled binary. (server)
    '''
    
    try:
        source_dir = download_source_code()
        if source_dir is None:
            return None

        os.chdir(source_dir)
        try:
            subprocess.run(["cmake", "-S", ".", "-B", "build", "-DLLAMA_CUBLAS=ON"], check=True)
            
            subprocess.run(["cmake", "--build", "build", "--config", "Release"], check=True)
            
            binary_name = "server.exe" if os.name == 'nt' else "server" 
            binary_path = os.path.join(source_dir, "build", "bin", binary_name)
            if os.path.exists(binary_path):
                print(f"Compilation successful. Binary located at: {binary_path}")
                return binary_path
            else:
                print("Compilation failed. Unable to find the binary file.")
                return None
                
        except subprocess.CalledProcessError as e:
            print(f"Compilation failed with error - input cmd `{e.cmd}`: {e}")
            print(f"Error output: {e.stderr} - {e.output}")
            return None
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            return None
    except Exception as e:
        print(f"An error occurred: {e}")
        return None


from PIserver.clients.FileDownloadClient import FileDownloadClient
from PIserver.commands.command import Command
from PIserver.setup import set_up
from PIserver.utils.files import *
from PIserver.constants import *
from pathlib import Path
from huggingface_hub import snapshot_download, HfApi
from tqdm import tqdm

class Clone_Model(Command):
    def register_subcommand(self, subparser):
        clone_parser = subparser.add_parser("clone", help="Download a model from https://powerinfer.com or update a local model.")
        clone_parser.add_argument("model", help="The model name to clone.")
        clone_parser.add_argument("local_dir", nargs='?', help="Clone the model into assigned local directory.")
        clone_parser.add_argument("-hf", help="Clone a runnable gguf model from huggingface.")
        
    def execute(self, args):
        if not Path(DEFAULT_STORAGE_PATH).exists():
            set_up()
        mname = args.model
        mname, tname = check_model_name_with_size(mname)
        
        # parse local path to store
        local_dir_name = str(args.model).replace(":", "-").replace("/", "-")
        local_path = Path(args.local_dir) if args.local_dir is not None else Path(DEFAULT_MODEL_PATH) / Path(local_dir_name)
        if args.local_dir is None:
            config = read_file(DEFAULT_CONFIG_FILE)
            local_path = Path(config["model_path"]) / Path(local_dir_name)
        local_path.mkdir(parents=True, exist_ok=True)
            
        if args.hf is not None:
            print("Cloning from huggingface...")
            try:
                api = HfApi()
                commits = api.list_repo_commits(args.hf)
                model_version = commits[0].commit_id[:8] if commits else "unknown"
                
                need_download = self.check_model(args.model, model_version)
                
                snapshot_download(
                    args.hf, 
                    local_dir=local_path, 
                    tqdm_class=tqdm
                )
                
                total_size = get_folder_size(local_path)
                add_row([mname, tname, total_size, model_version, str(local_path)])
                print(f"Successfully cloned model {args.model} into {str(local_path)}.")
            except KeyboardInterrupt:
                print("Download stopped.")
            return
        
        mname, uname = get_uname_from_model(mname)
        print(f"Trying to clone model {mname} size {tname}...")
        if tname == "" or uname == "":
            log_error("Invalid model name. Format the model name like 'USR/NAME:SIZE'.")
            return
        # check remote access
        getModelRequest = {
            "mname": mname,
            "tname": tname
        }
        if uname != "":
            getModelRequest["uname"] = uname
            
        try:
            client = FileDownloadClient()
            info = client.getModelInfo(getModelRequest)
            if info is None:
                return
            structure = info["dir_info"]
            print("Model info: ", info)
            print("structure: ", structure)
            need_download = self.check_model(args.model, info["version"])
            if not need_download:
                return
            # download & add to local model list    
            if client.download(local_path, structure):
                name = uname + "/" + mname if uname != "" else mname
                add_row([name, tname, info["size"], info["version"], str(local_path)])
        except KeyboardInterrupt:
            print("Download stopped.")
            return
        
    def check_model(self, mname, version) -> bool:
        # check in the list about version and path
        rows, rest = filter_rows(parse_condition(mname))
        # check remote model version(metadata)
        if len(rows) > 0:
            if version == rows[0][3]:
                print("The model is up to date. No need to download.")
                return False
            else: # delete old version
                print("Removing the older version of model...")
                remove_dir(rows[0][4])
                write_rows(rest)
        return True


import csv
from PIserver.constants import *
import tabulate
import shutil
from pathlib import Path
import sys
import json

# All the printings


def print_table(data: list, header: list):
    '''print data in table format using tabulate'''
    print(tabulate.tabulate(data, headers=header, tablefmt="plain"))
    
def log_error(msg):
    sys.stderr.write(f"Error: {msg}\n")

# All the file operations
def check_existence(path):
    return Path(path).exists()

def check_list_file():
    if not DEFAULT_MODEL_LIST_FILE.exists():
        with open(DEFAULT_MODEL_LIST_FILE, 'w') as f:
            writer = csv.writer(f)
            writer.writerow(LOCAL_LIST_HEADER)  

def get_absolute_path(path):
    return Path(path).resolve()

def get_folder_size(path: Path):
    return sum(f.stat().st_size if f.is_file() else get_folder_size(f) for f in path.glob('**/*'))

def remove_dir(path):
    path = Path(path)
    if not path.exists():
        log_error(f"Directory {path} not found.")
        return REMOVE_RESULT.NOT_FOUND
    try:
        shutil.rmtree(path)
    except Exception as e:
        log_error(f"Unable to remove directory {path}: {e}")
        return REMOVE_RESULT.ERROR  
    return REMOVE_RESULT.SUCCESS
        
def remove_file(path):
    path = Path(path)
    if path.exists():
        try:
            path.unlink()
            print(f"File {path} successfully removed.")
            return REMOVE_RESULT.SUCCESS
        except Exception as e:
            log_error(f"Unable to remove file {path}: {e}")
            return REMOVE_RESULT.ERROR
    else:
        return REMOVE_RESULT.NOT_FOUND
            
# All the csv file operations
def add_row(model: list):
    '''add a row to the model list file'''
    check_list_file()
    with open(DEFAULT_MODEL_LIST_FILE, 'a', newline='') as f:
        writer = csv.writer(f)
        writer.writerow(model)
    
def filter_rows(check):
    '''filter rows in the model list file'''
    check_list_file()
    with open(DEFAULT_MODEL_LIST_FILE, 'r') as f:
        reader = csv.reader(f)
        next(reader)
        rows, rest = [], []
        for row in reader:
            if len(row) > 0 and check(row):
                rows.append(row)
            else:
                rest.append(row)
        return rows, rest
    
def write_rows(rows: list):
    '''write rows to the model list file'''
    check_list_file()
    with open(DEFAULT_MODEL_LIST_FILE, 'w', newline='') as f:
        writer = csv.writer(f)
        writer.writerow(LOCAL_LIST_HEADER)
        writer.writerows(rows)

# All the string operations        
def parse_model(name):
    arr = name.split(':')
    return arr[0], arr[1] if len(arr) > 1 else None

def check_model_name_with_size(name):
    if ':' not in name:
        log_error("Please specify the model size. Format the model name like 'USR/NAME:SIZE'.")
        return name, ""
    return parse_model(name)

def get_uname_from_model(mname):
    if '/' not in mname:
        return mname, ""
    return mname.split('/')[1], mname.split('/')[0]

def parse_condition(model):
    name, size = parse_model(model)
    if size is None:
        return lambda x: x[0] == name
    else:
        return lambda x: x[0] == name and x[1] == size
    
# All the json file operations
def read_file(path: Path) -> dict:
    try:
        with open(path, 'r') as f:
            return dict(json.load(f))
    except json.JSONDecodeError:
        return {}
    except FileNotFoundError:
        path.touch(0o755, exist_ok=True)
        return {}
    
def write_file(path: Path | str, data: dict) -> bool:
    path = Path(path)
    try:
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, 'w') as f:
            json.dump(data, f, indent=4)
            return True
    except Exception as e:
        log_error(f"Unable to write to file {path}: {e}")
        return False
    
def add_engine(name, path):
        if not check_existence(path):
            log_error(f"File {path} does not exist. Please check the path and try again.")
            return False
        path = str(get_absolute_path(path))
        engines = read_file(DEFAULT_ENGINE_LIST_FILE)
        if name in engines:
            response = input(f"Engine {name} already exists. Do you want to overwrite it? (y/n)")
            if response != "y":
                print("Installation cancelled.")
                return False
            if engines[name] != path and check_existence(engines[name]):
                print(f"Removing old engine file {engines[name]}...")
                res = remove_file(engines[name])
                if res == REMOVE_RESULT.ERROR:
                    log_error(f"Unable to remove old engine file {engines[name]}. Remove it manually or try after fixing the error.")
                    return False
        engines[name] = path
        write_file(DEFAULT_ENGINE_LIST_FILE, engines)
        return True


# This script download installation packages according to user systems.

import curses
import platform

from PIserver.constants import ENGINE_CHOICES

def get_engine_choices():
    return ENGINE_CHOICES[platform.system()]

def get_engine_path(engine_name):
    return ENGINE_CHOICES[platform.system()][engine_name]

def display(stdscr, choices, selected, current_row):
    stdscr.clear()
    stdscr.addstr(0, 0, "Choose the installation packages (You can do it later using `powerinfer install` command):")
    stdscr.addstr(2, 0, "Press SPACE to select and ENTER to submit.")
    begin = 4
    for i, choice in enumerate(choices):
        prefix = "* " if i in selected else "  "
        if i == current_row:
            stdscr.addstr(i + begin, 0, prefix + choice, curses.A_REVERSE)
        else:
            stdscr.addstr(i + begin, 0, prefix + choice)
    stdscr.refresh()
    
def choosing(stdscr):
    choices = get_engine_choices()
    choice_names = list(choices.keys())
    
    curses.curs_set(0)
    curses.noecho()
    curses.cbreak()
    stdscr.keypad(True)
    
    selected = set() # TODO: load selected from engine list and choice
    current_row = 0
    while True:
        display(stdscr, choice_names, selected, current_row)
        key = stdscr.getch()

        if key == curses.KEY_UP and current_row > 0:
            current_row -= 1
        elif key == curses.KEY_DOWN and current_row < len(choices) - 1:
            current_row += 1
        elif key == ord(' '):
            if current_row in selected:
                selected.remove(current_row)
            else:
                selected.add(current_row)
        elif key == 10:
            for pkg in selected:
                single_install(choices[choice_names[pkg]])
            break
    return selected

def single_install(pkg_path) -> str:
    # return downloaded path
    print(pkg_path)

def interactive_install():
    print("Installing the backend engine...")
    curses.wrapper(choosing)


from PIserver.commands.command import Command
from PIserver.setup import set_up
from PIserver.utils.files import *

class Remove_Models(Command):
    def register_subcommand(self, subparser):
        rm_parser = subparser.add_parser("remove", help="Remove selected local model or local engine.")
        rm_parser.add_argument("model", nargs='?', help="The model name or engine name to remove.")
        rm_parser.add_argument("-i","--install", default=None, help="Remove the selected backend engine package.", action="store_true")
        
    def execute(self, args):
        if not Path(DEFAULT_STORAGE_PATH).exists():
            set_up()
        if args.install is not None:
            if args.model is None:
                log_error("Please specify the engine name to remove.")
                return
            engines = read_file(DEFAULT_ENGINE_LIST_FILE)
            if args.model not in engines:
                log_error(f"Engine {args.model} not found. Please check the name or use `powerinfer list -i` to see all the available engines.")
                return
            print(f"Removing engine {args.model}...")
            res = remove_file(engines[args.model])
            if res == REMOVE_RESULT.ERROR:
                log_error(f"Unable to remove engine {args.model}. Remove it manually or try after fixing the error.")
                return
            if res == REMOVE_RESULT.NOT_FOUND:
                log_error(f"Unable to find the engine file {args.model}.")
            # delete record from config file
            del engines[args.model]
            write_file(DEFAULT_ENGINE_LIST_FILE, engines)
            
        else:
            rows, rest = filter_rows(lambda x: True if args.model is None else parse_condition(args.model))
            if args.model is None and len(rows) > 0:
                response = input(f"Are you sure you want to remove all the local models? (y/n)")
                if response != "y":
                    return
            
            if args.model is not None:
                if len(rows) == 0:
                    log_error(f"Unable to find model: {args.model} locally. Please check your models using `pwi list`.")
                    return
                elif len(rows) > 1:
                    print("Found models:")
                    print_table(rows, LOCAL_LIST_HEADER)
                    print()
                    response = input(f"Found multiple models with name {args.model}. Do you want to remove all of them? (y/n)")
                    if response != 'y':
                        return
            
            for row in rows:
                res = remove_dir(row[4])
                if res == REMOVE_RESULT.ERROR:
                    log_error(f"Unable to remove model {row[0]}:{row[1]}.")
                    return
                write_rows(rest)
                print("Model successfully removed.")
        
        


from PIserver.commands.command import Command
from PIserver.constants import *
from PIserver.setup import set_up
from PIserver.utils.files import *


class List_Models(Command):
    def register_subcommand(self, subparser):
        list_parser = subparser.add_parser("list", help="Show local models. Show backend engines.")
        list_parser.add_argument("model", nargs='?', default=None, help="The model name to list. (Optional)")
        list_parser.add_argument("-i","--install", default=None, help="List all the backend engine packages.", action="store_true")
        
    def execute(self, args):
        if not Path(DEFAULT_STORAGE_PATH).exists():
            set_up()
        mname = args.model
        if args.install is not None:
            if args.model is not None:
                log_error("Cannot use -i and model together. Please use it seperately to list backend engines and models.")
                return
            print("Backend Engines:")
            engines = read_file(DEFAULT_ENGINE_LIST_FILE)
            print_table([[k, v] for k, v in engines.items()], ["ENGINE NAME", "ENGINE PATH"])
            return
        else:
            print("Local Models:")
            rows, rest = filter_rows(lambda x: x[0] != LOCAL_LIST_HEADER[0] if mname is None else parse_condition(mname))
            print_table(rows, LOCAL_LIST_HEADER)



import time
import json
import requests

class LLMClient:
    def __init__(self, host):
        self.host = host
        self.response = None
    
    def make_resData_stream(self, data, chat=False, time_now = 0, start=False):
        resData = {
            "id": "chatcmpl" if (chat) else "cmpl",
            "object": "chat.completion.chunk" if (chat) else "text_completion.chunk",
            "created": time_now,
            "model": "LLaMA_CPP",
            "choices": [
                {
                    "finish_reason": None,
                    "index": 0
                }
            ]
        }
        # slot_id = data["slot_id"]
        if (chat):
            if (start):
                resData["choices"][0]["delta"] =  {
                    "role": "assistant"
                }
            else:
                resData["choices"][0]["delta"] =  {
                    "content": data["content"]
                }
                if (data["stop"]):
                    resData["choices"][0]["finish_reason"] = "stop" if (data["stopped_eos"] or data["stopped_word"]) else "length"
        else:
            resData["choices"][0]["text"] = data["content"]
            if (data["stop"]):
                resData["choices"][0]["finish_reason"] = "stop" if (data["stopped_eos"] or data["stopped_word"]) else "length"

        return resData
    
    def generate(self, data, raw: bool=False):
        self.response = requests.post(self.host, json.dumps(data), stream=True)
        time_now = int(time.time())
        resData = self.make_resData_stream({}, chat=True, time_now=time_now, start=True)
        yield '{}\n'.format(json.dumps(resData))
        for line in self.response.iter_lines():
            if line:
                decoded_line = line.decode('utf-8')
                if raw:
                    yield decoded_line
                    continue
                resData = self.make_resData_stream(json.loads(decoded_line[6:]), chat=True, time_now=time_now)
                yield '{}\n'.format(json.dumps(resData))
                
    def close(self):
        self.response.close()
    
class PromptManager():
    def __init__(self, system_prompt):
        self.chat = []
        self.instruction = system_prompt
        
    def format_prompt(self, question):
        chat_history = ""
        for msg in self.chat:
            chat_history += f"User: {msg['user']}\nAssistant: {msg['assistant']}\n"
        return f"{self.instruction}{chat_history}User: {question}\nAssistant: "
                 
    def save_dialog(self, question, answer):
        self.chat.append({
            "user": question,
            "assistant": answer
        })


import requests
from PIserver.constants import *
from PIserver.utils.files import log_error
from typing import Optional
import json
from pathlib import Path
import time
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import serialization
import base64

def getHeader():
    pubkey = Path(DEFAULT_SSH_PUB_KEY_PATH).read_text()
    timestamp = str(int(time.time() * 1000))
    
    # get private key
    with open(DEFAULT_SSH_PEM_KEY_PATH, "rb") as key_file:
        private_key = serialization.load_pem_private_key(
            key_file.read(),
            password=None
        )

    signature = private_key.sign(
        timestamp.encode('utf-8'),
        padding.PKCS1v15(),
        hashes.SHA256()
    )
    signature_b64 = base64.b64encode(signature).decode('utf-8')
    return {"pubkey": pubkey, "timestamp": timestamp, "signature": signature_b64, "Content-Type": "application/json"}
    
def send_post_request(url: str, data=None, stream: bool = False, header: dict=None, params:dict = None) -> Optional[requests.Response]:
    try:
        headers = getHeader()
        if header is not None:
            headers.update(header)
        param = {} if params is None else params
        response = requests.post(
            backend_host+url, 
            data=json.dumps(data if data is not None else {}), 
            headers=headers, 
            stream=stream, 
            params=param
        )
        return response
    except Exception as e:
        log_error(f"Cannot successfully get response. {e}")
        return None
    
# def get_user_by_pub_key(pub_key: str) -> str:
#     # uid = send_get_request("/key/getUser", {"pub_key": pub_key})
#     # print(uid)
#     return ""


from PIserver.constants import *

LOCAL_LIST_HEADER = ['MODEL_NAME', 'SIZE', 'BSIZE', 'VERSION', 'PATH']

def generate_ssh_key():
    try:
        import paramiko
        key = paramiko.RSAKey.generate(2048)
        key.write_private_key_file(DEFAULT_SSH_PEM_KEY_PATH)
        pub = key.get_base64()
        with open(DEFAULT_SSH_PUB_KEY_PATH, 'w') as f:
            f.write("ssh-rsa "+pub)
    except ImportError:
        print("paramiko is not installed. Skipping SSH key generation.")

def generate_model_list_file():
    import csv
    DEFAULT_MODEL_LIST_FILE.touch(0o755, exist_ok=True)
    with open(DEFAULT_MODEL_LIST_FILE, 'w') as f:
        writer = csv.writer(f)
        writer.writerow(LOCAL_LIST_HEADER)
        
def generating_config_file():
    import json
    DEFAULT_CONFIG_FILE.touch(0o755, exist_ok=True)
    with open(DEFAULT_CONFIG_FILE, 'w') as f:
        json.dump(DEFAULT_CONFIG, f, indent=4)

def set_up():
    DEFAULT_STORAGE_PATH.mkdir(0o755, parents=True, exist_ok=True)
    DEFAULT_MODEL_PATH.mkdir(0o755, parents=True, exist_ok=True)
    generate_ssh_key() # FIXME: add ssh-key in production; in test don't change
    generate_model_list_file()
    
    DEFAULT_INSTALL_PATH.mkdir(0o755, parents=True, exist_ok=True)
    DEFAULT_ENGINE_LIST_FILE.touch(0o755, exist_ok=True)
    
    generating_config_file()


from setuptools import setup

setup(
    name='powerinfer-server',
    version='1.1.1',
    author='ARL',
    install_requires=[
        "fastapi>=0.68.0",
        "uvicorn>=0.15.0",
        "paramiko>=2.11.0",
        "requests>=2.25.0",
        "tqdm>=4.61.0",
        "tabulate>=0.8.9",
        "huggingface_hub>=0.10.1",
        "cryptography>=39.0.0",
        "keyboard>=0.13.5",
    ],
    extras_require={
        ':sys_platform == "win32"': [
            "windows-curses>=2.2.1",
        ],
    },
)


from pathlib import Path
from PIserver.commands.command import Command
from PIserver.constants import DEFAULT_CONFIG, DEFAULT_CONFIG_FILE, DEFAULT_STORAGE_PATH
from PIserver.setup import set_up
from PIserver.utils.files import read_file, write_file, log_error

class Config(Command):
    def register_subcommand(self, subparser):
        cfg_parser = subparser.add_parser("config", help="Manage the store location of models.")
        group = cfg_parser.add_mutually_exclusive_group()
        group.add_argument("-l","--list", default=None, help="Show current configurations", action="store_true")
        group.add_argument("-r","--reset", default=None, help="Change the config back into default values.", action="store_true")
        group.add_argument("-s","--set", default=None, help="Set the config in the format option=value.")
        group.add_argument("-rm", "--remove", default=None, help="Remove a config option.")
        group.add_argument("-cp", "--copy", default=None, help="Copy a config file into a new location.")
    
    def execute(self, args):
        if not Path(DEFAULT_STORAGE_PATH).exists():
            set_up()
        if args.list is not None:
            config = read_file(DEFAULT_CONFIG_FILE)
            print("Current configurations:")
            for key in config:
                print(f"{key}={config[key]}")
        elif args.reset is not None:
            config = DEFAULT_CONFIG
            write_file(DEFAULT_CONFIG_FILE, config)
            print("Config reset to default.")
        elif args.set is not None:
            config = read_file(DEFAULT_CONFIG_FILE)
            if '=' not in str(args.set):
                log_error("Invalid format. Please use option=value.")
                return
            [option, value] = str(args.set).split('=')
            config[option] = value
            write_file(DEFAULT_CONFIG_FILE, config)
            print(f"Config set: {option}={value}")
        elif args.remove is not None:
            config = read_file(DEFAULT_CONFIG_FILE)
            option = str(args.remove)
            if option in config:
                del config[option]
                print(f"Option {option} successfully removed.")
            else:
                log_error(f"Option {option} not found.")
        elif args.copy is not None:
            config = read_file(DEFAULT_CONFIG_FILE)
            new_path = str(args.copy)
            success = write_file(new_path, config)
            if success:
                print(f"Config file copied to {new_path}.")


from pathlib import Path
from enum import Enum

POWERINFER_HOST = "192.168.1.82"
POWERINFER_MODEL_HOST = POWERINFER_HOST # TODO: keep the same with backend host for now
POWERINFER_LOCAL_MODEL_HOST = "http://127.0.0.1:8080/completion"
POWERINFER_SERVER_PORT = 8088
backend_host = "http://" + POWERINFER_HOST + ":" + str(POWERINFER_SERVER_PORT)

DEFAULT_STORAGE_PATH = Path.home() / ".powerinfer"
DEFAULT_CONFIG_FILE = Path.home() / ".powerinfer" / "config.json"
DEFAULT_SSH_PEM_KEY_PATH = DEFAULT_STORAGE_PATH / "id_rsa"
DEFAULT_SSH_PUB_KEY_PATH = DEFAULT_STORAGE_PATH / "id_rsa.pub"
DEFAULT_MODEL_PATH = DEFAULT_STORAGE_PATH / "models"
DEFAULT_MODEL_LIST_FILE = DEFAULT_MODEL_PATH / "models.csv"
DEFAULT_INSTALL_PATH = DEFAULT_STORAGE_PATH / "engines"
DEFAULT_ENGINE_LIST_FILE = DEFAULT_INSTALL_PATH / "list.json"
TEST_SSH_PATH = Path.home() / ".ssh" / "id_rsa"

LOCAL_LIST_HEADER = ['MODEL_NAME', 'SIZE', 'BSIZE', 'VERSION', 'PATH']
REMOTE_LIST_HEADER = ['MODEL_NAME', 'ARCH', 'DOWNLOADS', 'LAST_UPDATED']
REMOTE_MODEL_TYPE_HEADER = ['SIZE', 'BSIZE', 'VERSION']

CHUNK_SIZE = 1024 * 1024 * 10  # 10MB per chunk

# Engine Choices
WINDOWS_ENGINE_CHOICES = {
    "windows-cpu-x64-843195e": "",
    "windows-cpu-x86": "",
    "windows-cuda-x64": "",
    "windows-cuda-x86": "",

}

LINUX_ENGINE_CHOICES = {

}

MAC_ENGINE_CHOICES = {}

ENGINE_CHOICES = {
    "Windows": WINDOWS_ENGINE_CHOICES,
    "Linux": LINUX_ENGINE_CHOICES,
    "Darwin": MAC_ENGINE_CHOICES
}

DEFAULT_ENGINE_NAME = "default-cuda"

DEFAULT_CONFIG = {
    "model_path": str(DEFAULT_MODEL_PATH),
    # "engine": str(DEFAULT_STORAGE_PATH / "PowerInfer" / "build" / "bin" / "server"),
    "options": {
        "n_predict": 512,
        "top_k": 40,
        "top_p": 0.9,
        "min_p": 0.05,
        "temp": 0.7,
        "stop": ["\nUser: "]
    },
    "system-prompt": "Transcript of a never ending dialog, where the User interacts with an Assistant.\nThe Assistant is helpful, kind, honest, good at writing, and never fails to answer the User's requests immediately and with precision.\nUser: Hi, Who are you?\n Assistant: I'm a helpful AI assistant. How can I help you today?\n",
}

class REMOVE_RESULT(Enum):
    SUCCESS = 0
    NOT_FOUND = 1
    ERROR = 2
    
class LLMState(Enum):
    RUNNING = 0
    STOPPED = 1
    ERROR = 2